-----------------------------------------------------------------------
-- Locals
-----------------------------------------------------------------------

local L = AceLibrary("AceLocale-2.2"):new("Violation")

local tablet = AceLibrary("Tablet-2.0")
local BC = AceLibrary("Babble-Class-2.2")
local roster = AceLibrary("Roster-2.1")
local AceOO = AceLibrary("AceOO-2.0")
local dew = AceLibrary("Dewdrop-2.0")


local ChatThrottleLib = ChatThrottleLib
if type(ChatThrottleLib) ~= "table" or type(ChatThrottleLib.version) ~= "number" or ChatThrottleLib.version < 13 then
	error("Violation requires ChatThrottleLib 13 or higher.")
end

local SM = LibStub("LibSharedMedia-2.0")

local coloredNames = setmetatable({}, {__index =
	function(self, key)
		local obj = roster:GetUnitObjectFromName(key)
		if not obj then return key end
		self[key] = format("|cff%s%s|r", BC:GetHexColor(obj.class), key)
		return self[key]
	end
})

local RGB = setmetatable({}, {__index =
	function(self, key)
		local obj = roster:GetUnitObjectFromName(key)
		if not obj then return nil end
		local r, g, b = BC:GetColor(obj.class)
		self[key] = {r, g, b}
		return self[key]
	end
})

local db = nil

local _G = getfenv(0)
local table_insert = table.insert

local playerName = nil
local percentString = "(%.1f%%)"

local moduleVersion = 5
local latestRevisionSeen
local updateWarned

local runningModules = {}
local runningModuleOrdering = {}
local violationUsers = nil
local characterPets = {}
local moduleData = {}
local dataFormatter = {}
local characterDataSource = {}

local VIOLATION_LEVEL = 4       -- true violation user
local VIOLATION_PROXY_LEVEL = 3 -- proxy violation user meaning pets and people broadcasted by other vio users
local SWS_LEVEL = 2             -- SWS user
local DAMAGEMETERS_LEVEL = 1    -- DMM user
local COMBATLOG_LEVEL = 0       -- CombatLog Users

local red = "Interface\\AddOns\\Violation\\icons\\red"
local blue = "Interface\\AddOns\\Violation\\icons\\blue"
local grey = "Interface\\AddOns\\Violation\\icons\\grey"

local grouped = nil -- grouped status

local violationCategories = {
	damage  = 100,
	healing = 110,
	threat  = 120,
	other   = 130
}

-----------------------------------------------------------------------
-- Local heap and utility functions
-----------------------------------------------------------------------

local new, del
do
	local cache = setmetatable({},{__mode="k"})
	function new(...)
		local t = next(cache)
		if t then
			cache[t] = nil
			for i = 1, select("#", ...) do
				table_insert(t, (select(i, ...)))
			end
			return t
		else
			return {...}
		end
	end
	function newkeyed(...)
		local t = next(cache)
		if t then
			cache[t] = nil
		else
			t = {}
		end
		for i = 1, select("#", ...), 2 do
			t[select(i, ...)] = select(i+1, ...)
		end
		return t
	end
	function del(t)
		if type(t) == "table" then
			for k, v in pairs(t) do
				if type(v) == "table" then
					v = del(v)
				end
				t[k] = nil
			end
			cache[t] = true
		end
		return nil
	end
end

local function Colorize(text, r, g, b)
	return string.format("|cff%02x%02x%02x%s|r", r * 255, g * 255, b * 255, text)
end


-----------------------------------------------------------------------
-- Addon declaration
-----------------------------------------------------------------------
Violation = AceLibrary("AceAddon-2.0"):new(
	"AceEvent-2.0",
	"AceDB-2.0",
	"AceConsole-2.0",
	"AceComm-2.0",
	"FuBarPlugin-2.0",
	"AceModuleCore-2.0"
)
local Violation = Violation

Violation:SetModuleMixins("AceEvent-2.0", "Parser-3.0")
Violation.hasIcon = true
Violation.independentProfile = true
Violation.hideWithoutStandby = true

-----------------------------------------------------------------------
-- Options
-----------------------------------------------------------------------

local function get(key)
	return db[key]
end
local function set(key, val)
	db[key] = val
end

local reportMethod = "SAY"
local reportTarget = ""

local reportToMenu = {
	whisper = {
		type = "group",
		name = L["Whisper"],
		desc = L["Whisper the data to someone."],
		pass = true,
		args = {
			player = {
				type = "text",
				name = "Player",
				desc = "Whisper the data to the specified player.",
				usage = "<playerName>",
				order = 1,
			},
			spacer = {
				type = "header",
				name = " ",
				order = 50,
			},
		},
		isChecked = function() return reportMethod=="WHISPER" end,
		order = 1,
	},
	say = {
		type = "toggle",
		name = L["Say"],
		desc = L["Print module data to say."],
		passValue = {method="SAY"},
		order = 2,
	},
	guild = {
		type = "toggle",
		name = L["Guild"],
		desc = L["Print module data on guild channel."],
		passValue = {method="GUILD"},
		order = 2,
	},
	officer = {
		type = "toggle",
		name = L["Officer"],
		desc = L["Print module data on officer channel."],
		passValue = {method="OFFICER"},
		order = 2,
	},
	party = {
		type = "toggle",
		name = L["Party"],
		desc = L["Print module data on party channel."],
		passValue = {method="PARTY"},
		order = 2,
	},
	raid = {
		type = "toggle",
		name = L["Raid"],
		desc = L["Print module data on raid channel."],
		passValue = {method="RAID"},
		order = 2,
	},
	cspacer = { -- space before the custom channels
		type = "header",
		name = " ",
		order = 50,
	},
	spacer = { -- spacer after the custom channels
		type = "header",
		name = " ",
		order = 90,
	},
	limit = {
		type = "range",
		name = L["Lines"],
		desc = L["Maximum number of lines to print."],
		max = 40,
		min = 1,
		step = 1,
		get = get,
		set = set,
		passValue = "maxPrintLines",
		order = 91,
	},
}

local options = {
	type = "group",
	handler = Violation,
	args = {
		track = {
			type = "group",
			name = L["Track"],
			desc = L["What data types to track."],
			disabled = "~IsActive",
			args = {}, -- Auto-populated by module toggles.
			order = 101,
			pass = true,
			get = function(mod)
				return db.enabledModules[mod]
			end,
			set = "ToggleModule",
		},
		fubar = {
			type = "group",
			name = L["Text"],
			desc = L["What data to show on the FuBar panel, if available."],
			disabled = function() return not FuBar end,
			args = {}, -- Auto-populated by module toggles.
			order = 102,
			pass = true,
			get = function(mod)
				return db.textModules[mod]
			end,
			set = function(mod, val)
				db.textModules[mod] = val
				Violation:UpdateText()
			end,
		},
		colors = {
			type = "group",
			name = L["Colors"],
			desc = L["Assign colors to the different data types."],
			pass = true,
			get = function(mod)
				if mod == "tooltip" then
					return db.colorTooltip
				else
					local tuple = db.colors[mod] or {r=1,g=1,b=1}
					return tuple.r or 1, tuple.g or 1, tuple.b or 1
				end
			end,
			set = function(mod, r, g, b)
				if mod == "tooltip" then
					db.colorTooltip = r
					Violation:UpdateTooltip()
				else
					local tuple = db.colors[mod]
					tuple.r = r
					tuple.g = g
					tuple.b = b
					Violation:ScheduleEvent("ColorUpdate", Violation.UpdateDisplay, 5, Violation)
					Violation:UpdateWindows(true)
				end
			end,
			args = {
				tooltip = {
					type = "toggle",
					name = L["Color the tooltip"],
					desc = L["Toggles whether or not to use these colors in the tooltip."],
					order = 1,
					disabled = true,
				},
				spacer = {
					type = "header",
					name = " ",
					order = 50,
				},
			}, -- Auto-populated by module toggles.
			order = 103,
		},
		tooltip = {
			type = "group",
			name = L["Tooltip"],
			desc = L["What data to show in the tooltip."],
			args = {},
			order = 104,
			pass = true,
			get = function(mod)
				return db.tooltipModules[mod]
			end,
			set = function(mod, val)
				db.tooltipModules[mod] = val
				Violation:UpdateTooltip()
			end
		},
		windowspacer = {
			type = "header",
			name = " ",
			order = 110,
		},
		windows = {
			type = "group",
			name = L["Windows"],
			desc = L["Window Options"],
			args = {
				spacer1 = {
					type = "header",
					text = " ",
					order = 110,
				},
				newwindow = {
					type = "execute",
					name = L["Create New Window"],
					desc = L["Create a new bar window"],
					func = function()
						db.windows.count = db.windows.count + 1
						Violation:SetupWindows()
					end,
					disabled = function() return not next(runningModules) end,
					order = 111,
				},
				show = {
					type = "toggle",
					name = L["Show Windows"],
					desc = L["Toggle showing all windows on/off"],
					set = function(v) Violation:ToggleWindows(v) end,
					get = function() return db.showWindows end,
					order = 112,
				},
				showgrouped = {
					type = "toggle",
					name = L["Show in group"],
					desc = L["Show Windows when in a group."],
					set = function(v)
						db.showGrouped = v
						if v then
							if grouped then
								Violation:ToggleWindows(true)
							else
								Violation:ToggleWindows(false)
							end
						else
							Violation:ToggleWindows(true)
						end
					end,
					get = function() return db.showGrouped end,
					order = 113,
				},
				spacer2 = {
					type = "header",
					text = " ",
					order = 114,
				},
				windowdefaulttexture = {
					type = "text",
					name = L["Default Texture"],
					desc = L["Default Bar Texture"],
					set = function(v)
						db.defaultBarTexture = v
						Violation:UpdateWindows()
					end,
					get = function() return db.defaultBarTexture end,
					validate = SM:List("statusbar"),
					order = 115,
				},
				windowtooltip = {
					type = "toggle",
					name = L["Show window bar tooltips"],
					desc = L["Enable showing tooltips on window bar hovering.\nWhen enabled, hold control to force display while in combat."],
					set = function(v) db.windowTooltips = v end,
					get = function() return db.windowTooltips end,
					order = 116,
				},
			},
			order = 111,
		},
		mergespacer = {
			type = "header",
			name = " ",
			order = 120,
		},
		merge = {
			type = "toggle",
			name = L["Merge pets"],
			desc = L["Merge pet data with master."],
			get = function() return db.mergePets end,
			set = function(value)
				db.mergePets = value
				Violation:UpdateDisplay()
				Violation:UpdateWindows()
			end,
			order = 121,
		},
		resetspacer = {
			type = "header",
			name = " ",
			order = 150,
		},
		resetPrint = {
			type = "toggle",
			name = L["Print on reset"],
			desc = L["Prints your current data locally when your data is reset."],
			get = get,
			set = set,
			passValue = "printOnReset",
			order = 200,
		},
		reset = {
			type = "group",
			name = L["Reset"],
			desc = L["Options for when to reset the data."],
			args = {
				now = {
					type = "execute",
					name = L["Reset now"],
					desc = L["Immediately resets your data."],
					func = "Reset",
					order = 100,
				},
				group = {
					type = "execute",
					name = L["Reset group"],
					desc = L["Resets the Violation data for every group member. Only the group leader or an officer can do this."],
					func = "ResetGroup",
					disabled = function()
						return not IsRaidLeader() and not IsRaidOfficer() and not IsPartyLeader()
					end,
					order = 101,
				},
				leader = {
					type = "toggle",
					name = L["Leader resets"],
					desc = L["Accept reset requests from raid leaders or officers."],
					get = get,
					set = set,
					passValue = "leaderReset",
					order = 102,
				},
				spacer = {
					type = "header",
					name = " ",
					order = 150,
				},
				autoheader = {
					type = "header",
					name = L["Auto-reset"],
					order = 200,
				},
				gaincombat = {
					type = "toggle",
					name = L["Entering combat"],
					desc = L["Automatically reset the data when entering combat."],
					get = function() return db.gainCombat end,
					set = function(v)
						if v then
							Violation:RegisterEvent("PLAYER_REGEN_DISABLED", "Reset")
						else
							Violation:UnregisterEvent("PLAYER_REGEN_DISABLED")
						end
						db.gainCombat = v
					end,
					order = 201,
				},
				leavecombat = {
					type = "toggle",
					name = L["Leaving combat"],
					desc = L["Automatically reset the data when leaving combat."],
					get = function() return db.leaveCombat end,
					set = function(v)
						if v then
							Violation:RegisterEvent("PLAYER_REGEN_ENABLED", "Reset")
						else
							Violation:UnregisterEvent("PLAYER_REGEN_ENABLED")
						end
						db.leaveCombat = v
					end,
					order = 202,
				},
			},
			order = 201,
		},
		report = {
			type = "group",
			name = L["Report"],
			desc = L["Report your current data to a channel."],
			args = {},
			order = 202,
		},
		spying = {
			type = "group",
			name = L["Spying"],
			desc = L["Spying allows Violation to read data sent by other meter addons. "]
					.. Colorize(L["Although it receives their data, it will not send anything back to them."],1,0.5,0),
			order = 203,
			args = {
				dmm = {
					type = "toggle",
					name = L["DamageMeters"],
					desc = L["Spy DamageMeters."],
					get = function() return db.spying.damageMeters end,
					set = function(v)
						db.spying.damageMeters = v
						Violation:UpdateSpying()
					end,
				},
				sws = {
					type = "toggle",
					name = L["SWStats"],
					desc = L["Spy SWStats."],
					get = function() return db.spying.swStats end,
					set = function(v)
						db.spying.swStats = v
						Violation:UpdateSpying()
					end,
				},
			},
		},
		syncInfo = {
			type = "execute",
			name = L["Syncing report"],
			desc = L["Display which player are actually sending data."],
			func = "ReportUsers",
			order = 204,
		},
	},
}

-----------------------------------------------------------------------
-- Module prototype
-----------------------------------------------------------------------

Violation.modulePrototype.version = -1
Violation.modulePrototype.canBroadcast = true

local incompatible = {}
local registered = {}
function Violation.modulePrototype:OnInitialize()
	if type(self.name) ~= "string" then
		error("A module is missing its name field.")
	elseif type(self.version) ~= "number" then
		error(self.name .. " doesn't have a version field.")
	elseif registered[self.name] then
		error("A module has been initialized twice for some reason.")
	end

	if self.version == moduleVersion then
		Violation:OnRegister(self)
		registered[self.name] = true
	else
		incompatible[self] = true
		Violation:ModuleIncompatible(self)
	end
end

function Violation.modulePrototype:OnEnable(first)
	if not db then db = Violation.db.profile end
	if incompatible[self] or not db.enabledModules[self.name] then return end

	if not runningModules[self.name] then
		-- Always reset our data when we enable.
		runningModules[self.name] = self

		self.playerName = playerName
		self:Reset()

		if type(self.OnModuleEnable) == "function" then
			self:OnModuleEnable()
		end

		Violation:TriggerEvent("Violation_ModuleEnabled", self.name)

		Violation:UpdateWindows()
	end
end

function Violation.modulePrototype:OnDisable()
	if runningModules[self.name] then
		if type(self.OnModuleDisable) == "function" then
			self:OnModuleDisable()
		end

		local availableMod = next(runningModules, self.name)

		runningModules[self.name] = nil

		Violation:UpdateWindows()

		Violation:TriggerEvent("Violation_ModuleDisabled", self.name)
	end
end

function Violation.modulePrototype:AcceptDataFor(name)
	return Violation:AcceptDataFromSource(name, self.name)
end

function Violation.modulePrototype:GetSourceVictimAmount(info)
	return info.sourceName, info.recipientName, (info.amount or 0)
end

function Violation.modulePrototype:IsLocal(name)
	return name == playerName or characterDataSource[name] == playerName
end

function Violation.modulePrototype:GetRevision()
	return self.revision
end

function Violation.modulePrototype:Reset()
	if type(self.data) == "table" then
		self.data = del(self.data)
	end
	self.data = new()
	self.data[playerName] = 0
	if type(self.broadcast) == "table" then
		self.broadcast = del(self.broadcast)
	end
	self.broadcast = new()
	self.dirty = nil
	if type(self.OnReset) == "function" then
		self:OnReset()
	end
end

function Violation.modulePrototype:IsDirty()
	return self.dirty
end

function Violation.modulePrototype:ResetDirty()
	self.dirty = nil
end


function Violation.modulePrototype:IsBroadcastDirty(charName)
	if not charName then
		charName = playerName
	end
	return self.broadcast[charName] and true or false
end

function Violation.modulePrototype:GetBroadcastData(charName)
	if not charName then
		charName = playerName
	end
	local x = self.broadcast[charName]
	self.broadcast[charName] = nil
	return x
end

function Violation.modulePrototype:GetData(name)
	if name then
		return self.data[name]
	else
		return self.data
	end
end

function Violation.modulePrototype:IsShown()
	local name = self.name
	if runningModules[name] then
		if db.textModules[name] or db.tooltipModules[name] then
			return true
		end
		for k, w in ipairs(Violation.windows) do
			if w.options.module == name then
				return true
			end
		end
	end
	return false
end

function Violation.modulePrototype:RequestDisplayRefresh()
	if self:IsShown() then
		self:TriggerEvent('Violation_UpdateDisplay', self.name)
	end
end

function Violation.modulePrototype.dataFormatter(v)
	if not v then return "" end
	return tostring(v)
end

function Violation.modulePrototype:OnRemoteData(name, input)
	self:OnData(name, input, true)
end

function Violation.modulePrototype:new(...) return new(...) end
function Violation.modulePrototype:del(t) return del(t) end

function Violation.modulePrototype:setReporting(a,b,c)	-- called from aceoptions
	if type(a)=="table" then
		reportMethod = a.method
		reportTarget = a.target or reportTarget
	end
	if a=="player" then
		reportMethod = "WHISPER"
		reportTarget=b
	end
	if reportMethod=="WHISPER" then
		reportToMenu.whisper.name = format("%s %s", L["Whisper"], reportTarget)
	end
	self:onClickReporting()
end

function Violation.modulePrototype:getReporting(a,b,c) -- called from aceoptions
	if type(a)=="table" then
		if a.method=="WHISPER" then
			return a.target == reportTarget
		end
		return a.method==reportMethod and (a.target==nil or a.target==reportTarget)
	end
	if a=="player" then
		return reportTarget
	end
end

function Violation.modulePrototype:onClickReporting()	-- called from aceoptions
	Violation:PrintModuleDataToChannel(self.name, reportMethod, reportTarget)
end

-----------------------------------------------------------------------
-- Default module data implementation
-----------------------------------------------------------------------
-- This implmentation supposes that the modules handle "incremental" data
-- like damage done, or healing taken, ...
-- Special modules like DPS or HPS, have to override theses methods

Violation.modulePrototype.canMergeData = true

function Violation.modulePrototype:OnData(name, input, remote)
	if name and input and input > 0 then
		if not remote then
			self.broadcast[name] = (self.broadcast[name] or 0) + input
		end
		self.data[name] = (self.data[name] or 0) + input
		self.dirty = true
	end
end

function Violation.modulePrototype:GetMergedData(...)
	local sum = 0
	for i = 1,select("#", ...) do
		sum = sum + (self:GetData(select(i, ...)) or 0)
	end
	return sum
end

-----------------------------------------------------------------------
-- Implicit module interface
-----------------------------------------------------------------------

local AceOO = AceLibrary("AceOO-2.0")
local IViolation = AceOO.Interface{

	-- Optional methods:
	--
	-- -- Called when the module is enabled by the core. Any event registration
	-- -- should happen here.
	-- OnModuleEnable
	-- OnModuleDisable
	--
	-- -- Called when the module is initially registered.
	-- OnModuleRegister

	-- Must return the current module revision as a number.
	GetRevision = "function",

	-- Must return the module name as it should be displayed in the menu, for
	-- example "Damage done".
	GetDisplayName = "function",

	-- Must return the module name as it should be displayed in a condensed
	-- display, like "PlayerDPS" should return "DPS".
	GetShortDisplayName = "function",

	-- Must return whether the modules broadcastdata is dirty or not. If it is, a
	-- broadcast will be scheduled by the core.
	IsBroadcastDirty = "function",

	-- Must return wether the modules data is dirty or not. If it is a
	-- diplay update will be scheduled by the core.
	IsDirty = "function",

	-- This function is must reset the modules dirty flag that is returned by IsDirty.
	ResetDirty = "function",

	-- Must return the data that should be broadcasted to the other players.
	-- Must return 1 number only. This function should reset the broadcastdirty flag as well.
	GetBroadcastData = "function",

	-- Invoked when remote data is received for this module. Usually invoked by
	-- the core, but can also be invoked by other modules when they parse data
	-- from other addons, for example.
	-- Will only get two arguments, the first will be the person the data
	-- belongs to and the second will be a number.
	OnRemoteData = "function",

	-- Invoked by the core when all data should be reset.
	Reset = "function",

	-- Must return a universal unique ID that will be used for this module when
	-- sending data (where the universe is Violation)
	-- (Not used now, we use a memoization of the module name instead.
	-- GetUUID = "function",

	-- Must return the current dataset for this module. If an argument is given,
	-- it can be assumed to be a player name, and only the data (a number) for
	-- that player (or nil) must be returned.
	GetData = "function",
}

-----------------------------------------------------------------------
-- Addon initialization
-----------------------------------------------------------------------

local profileDefaults = {
	enabledModules = {
		["*"] = true,
	},
	textModules = {
		["Damage per Second"] = true,
	},
	tooltipModules = {
		["Damage per Second"] = true,
	},
	colors = {
		["**"]                 = { r = 1,   g = 1,   b = 1   },

		["Damage done"]        = { r = 0.8, g = 0.3, b = 0.3 },
		["Damage taken"]       = { r = 0.8, g = 0.5, b = 0.3 },
		["Damage per Second"]  = { r = 0.8, g = 0.8, b = 0.3 },
		["Healing done"]       = { r = 0.5, g = 0.8, b = 0.3 },
		["Healing taken"]      = { r = 0.3, g = 0.8, b = 0.3 },
		["Healing per Second"] = { r = 0.3, g = 0.8, b = 0.5 },
		["Overhealing done"]   = { r = 0.3, g = 0.8, b = 0.8 },
		["Global Threat"]      = { r = 0.3, g = 0.5, b = 0.8 },
		["Target Threat"]      = { r = 0.3, g = 0.3, b = 0.8 },
		["Decursing done"]     = { r = 0.5, g = 0.3, b = 0.8 },
		["Dispelling done"]    = { r = 0.0, g = 0.3, b = 0.8 },
		["Critical hit rate"]  = { r = 0.8, g = 0.3, b = 0.8 },
		["Aggro duration"]     = { r = 0.8, g = 0.3, b = 0.5 },
	},
	colorTooltip = true,
	gainCombat = nil,
	leaveCombat = nil,
	leaderReset = true,
	printOnReset = true,
	mergePets = true,
	windowTooltips = true,
	defaultBarTexture = "Blizzard",

	-- Spying
	spying = {
		damageMeters = false,
		swStats = false,
	},

	windows = {
		count = 1,
		["**"] = {
			width = 200,
			height = 200,
			nrBars = 10,
			scale = 1,
			show = true,
			module = "Damage done",
			locked = false,
			useModuleColor = true,
			color = { r = 1, g = 0, b = 0 },
			useDefaultTexture = true,
			barTexture = "Blizzard",
			useSimpleBars = false,
		},
	},

	showWindows = true,
}

function Violation:OnInitialize()
	local revision = tonumber(("$Revision: 57181 $"):sub(12, -3)) or 1
	if not self.version then self.version = "1.0" end
	self.version = self.version .. "." .. revision
	self.revision = revision

	_G.VIOLATION_VERSION = self.version
	_G.VIOLATION_REVISION = self.revision

	self:RegisterDB("ViolationDB", "ViolationPerCharDB", "Default")
	self:RegisterDefaults("profile", profileDefaults)

	self.OnMenuRequest = options

	self:RegisterChatCommand("/violation", self.OnMenuRequest, "VIOLATION")

	self:SetCommPrefix("Violation")
	self:SetDefaultCommPriority("BULK")

	self:SetIcon(grey)

	playerName = UnitName("player")

	-- exports
	self.options = options
	self.runningModules = runningModules
	self.dataFormatter = dataFormatter

	-- This guy used Vio back when it was hardcoded self.db.char?
	if self.db.char.windows and self.db.char.importedToProfile~=1 then
		-- Schedule a near-immediate event to switch profile.
		-- Can't be done immediately since that means we'd start seeing :OnProfileEnable() callbacks right here, when we're NOT ready for them
		self:ScheduleEvent(self.CopyOldProfile, 0.5, self)
	end

end

function Violation:CopyOldProfile()
	self:RegisterDefaults("char", profileDefaults) -- Force defaults into the old per-char db so we copy sane settings
	self:SetProfile("char")

	local function copyInto(dst,src)
		for k,v in pairs(dst) do
			if type(v)~="table" then
				dst[k]=nil
			end
		end
		for k,v in pairs(src) do
			if type(v)=="table" then
				if type(dst[k])~="table" then
					dst[k]={}
				end
				copyInto(dst[k], v)
			else
				dst[k] = v
			end
		end
	end
	copyInto(self.db.profile, self.db.char)

	self.db.char.importedToProfile = 1

	-- make violation pick up all the changes we copied in
	self:OnProfileDisable()
	self:OnProfileEnable()
end


function Violation:OnEnable(first)
	db = self.db.profile

	grouped = nil

	-- register these events before we try to do anything with LoD modules
	self:RegisterBucketEvent("Violation_ModuleEnabled", 0.2, "UpdateRunningModuleOrdering")
	self:RegisterBucketEvent("Violation_ModuleDisabled", 0.2, "UpdateRunningModuleOrdering")

	-- get the LoD modules that should be loaded before we register memoizations
	self:RegisterModulePacks()

	if first then
		local memoizations = new()
		for name in self:IterateModules() do
			table_insert(memoizations, name)
		end
		table_insert(memoizations, "LeaderReset")
		table_insert(memoizations, "ViolationPing")
		table_insert(memoizations, "ViolationPong")
		self:RegisterMemoizations(memoizations)
		memoizations = del(memoizations)

	end

	self:OnProfileEnable()

	self:RegisterComm("Violation", "GROUP", "OnComm")
	self:RegisterComm("Violation", "WHISPER", "OnPrivateComm")

	self:ScheduleRepeatingEvent("Broadcast", self.BroadcastAll, 5, self)
	self:ScheduleRepeatingEvent("DisplayUpdate", self.UpdateDisplayIfDirty, .5, self)	-- 2 is too long for threat, at least early in combat /Mikk

	if db.gainCombat then
		self:RegisterEvent("PLAYER_REGEN_DISABLED", "Reset")
	end

	if db.leaveCombat then
		self:RegisterEvent("PLAYER_REGEN_ENABLED", "Reset")
	end

	self:RegisterEvent("CHAT_MSG_CHANNEL_NOTICE", "UpdateChannels")

	self:RegisterBucketEvent("RosterLib_RosterUpdated", 2, function()
		self:SendCommMessage("GROUP", "ViolationPing")
		self:UpdatePartyMenu()
		self:OnRosterUpdate()
	end)

	self:RegisterBucketEvent("Violation_UpdateDisplay", 0.1, "UpdateDisplayIfDirty")

	self:SetIcon(red)
	self:OnRosterUpdate(true)
	self:UpdateChannels()
	self:Reset()
end

function Violation:OnProfileEnable()
	db = self.db.profile

	-- Enable/disable all modules.
	for name in self:IterateModules() do
		self:ToggleModuleActive(name, db.enabledModules[name])
	end

	self:UpdateSpying()

	self:SetupWindows()
	self:UpdateText()
	self:ToggleWindows(db.showWindows)
end

function Violation:OnProfileDisable()
	-- Delete all windows. Start from the highest numbered one or window:GetID() goes bonkers
	local i=0
	while self.windows[i+1] do
		i=i+1
	end
	for j=i,1,-1 do
		self.windows[j]:Delete()
		self.windows[j]=nil
	end
end


function Violation:OnDisable()
	-- Close all windows
	self:CloseWindows()

	-- Disable all modules.
	for name, mod in pairs(runningModules) do
		self:ToggleModuleActive(name, false)
	end

	self:SetIcon(grey)
end

local modulepacks
local moduleInPack

function Violation:InstallModulePack(pack, addonid, moduleMeta, packLocale, moduleMetaLocale)

	local providedModules = new(strsplit(",", moduleMeta))
	local providedModulesLocale
	if moduleMetaLocale then
		providedModulesLocale = new(strsplit(",", moduleMetaLocale))
	end

	if not modulepacks then
		modulepacks = new()
	end
	if not moduleInPack then
		moduleInPack = new()
	end
	modulepacks[pack] = new()
	modulepacks[pack].id = addonid
	modulepacks[pack].modules = providedModules

	local module,moduleLocale, goload
	for k, module in ipairs(providedModules) do
			if moduleInPack[module] then
				error(L["Module '%s' is already provided by '%s' for LoD."]:format(module, moduleInPack[module]))
			else
				moduleInPack[module] = pack
				if db.enabledModules[module] then
					goload = true
				end
				if providedModulesLocale then
					moduleLocale = providedModulesLocale[k]
				else
					moduleLocale = module
				end
				self:SetupModuleLoDMenu(module, moduleLocale)
			end
	end
	-- this pack has one or more modules enabled lets load it up
	if goload then
		self:LoadModulePack(pack)
	end
end

local function loadModule(moduleName)
	if moduleInPack[moduleName] then
		Violation:LoadModule(moduleName)
	end
end

function Violation:SetupModuleLoDMenu(moduleName, displayName)
	if not options.args.track.args.spacer then
		options.args.track.args.spacer = {
			type = "header",
			name = " ",
			order = 1001,
		}
	end
	options.args.track.args[moduleName] = {
		type = "execute",
		name = L["Load %s"]:format(displayName),
		desc = L["Toggle whether to track %q."]:format(displayName),
		func = "LoadModule",
		order = 1002,
		passValue = moduleName,
	}
end

function Violation:RemoveModuleLoDMenu(moduleName)
	options.args.track.args[moduleName] = nil
end

function Violation:RegisterModulePacks()
	local locale = GetLocale()
	for i=1,GetNumAddOns() do
		if not IsAddOnLoaded(i) and IsAddOnLoadOnDemand(i) then
			local addon = GetAddOnInfo(i)
			local pack = GetAddOnMetadata(addon, "X-Violation-Pack")
			if pack then
				local meta = GetAddOnMetadata(addon, "X-Violation-Modules")
				if meta then
					local packloc = GetAddOnMetadata(addon, "X-Violation-Pack-"..locale)
					local metaloc = GetAddOnMetadata(addon, "X-Violation-Modules-"..locale)
					if packloc then pack = packloc end
					if metaloc then meta = metaloc end
					-- feed localized data if able, this makes for pretty menu entries
					self:InstallModulePack(pack, i, meta, packloc, metaloc)
				end
			end
		end
	end
end

function Violation:LoadModule(moduleName)
	if moduleInPack[moduleName] then
		self:LoadModulePack(moduleInPack[moduleName])
		self:ToggleModule(moduleName, true)
	end
end

function Violation:LoadModulePack(pack)
	if modulepacks[pack] and not IsAddOnLoaded(modulepacks[pack].id) then
		-- remove the LoD Menu Entries
		for k, v in pairs(modulepacks[pack].modules) do
			self:RemoveModuleLoDMenu(v)
		end
		-- all we do is load the addon, the rest should go automagically
		LoadAddOn(modulepacks[pack].id)
		modulepacks[pack] = del(modulepacks[pack])
		for k, v in pairs(modulepacks) do
			return
		end
		options.args.track.args.spacer = nil
	end
end


function Violation:Reset()
	for index, modName in self:IterateRunningModules() do
		local module = runningModules[modName]
		if db.printOnReset then
			local data = self:GetModuleData(modName, playerName)
			if type(data) == "number" and data > 0 then
				local color = db.colors[modName]
				local name = module:GetDisplayName()
				local formatter = dataFormatter[modName] or tostring
				name = Colorize(name, color.r, color.g, color.b)
				data = formatter(data)
				self:Print(string.format("%s: %s", name, data))
			end
		end
		module:Reset()
	end
	self:UpdateDisplay()
	self:UpdateWindows()
end

local function checkCategory(cat)
	if not cat or not violationCategories[cat] then
		cat = "other"
	end
	local order = violationCategories[cat]
	return cat, order
end

function Violation:UpdateChannels()
	local id, name, cid, i
	for i=1, 10, 1 do
		id, name = GetChannelName(i)
		cid = "channel"..i
		if id > 0 and name then
			if not reportToMenu[cid] then
				reportToMenu[cid] = {
						type = "toggle",
						name = name,
						desc = L["Print module data on %s channel."]:format(name),
						passValue = {method = "CHANNEL", target = id},
						order = 50 + id,
				}
			else
				reportToMenu[cid].hidden = false
				reportToMenu[cid].name = name
				reportToMenu[cid].desc = L["Print module data on %s channel."]:format(name)
			end
		elseif reportToMenu[cid] then
			reportToMenu[cid].hidden = true
		end
	end
end

-----------------------------------------------------------------------
-- Module handling
-----------------------------------------------------------------------


function Violation:OnRegister(mod)
	if not AceOO.inherits(mod, IViolation) then
		error(string.format("The module %q does not inherit from the IViolation interface.", tostring(mod.name)))
	end

	if mod.revision and mod.revision > self.revision then
		self.revision = mod.revision
	end

	local moduleName = mod.name
	local displayName = mod:GetDisplayName()
	mod.violationCategory, mod.violationOrdering = checkCategory(mod.violationCategory)
	local order = mod.violationOrdering

	local function disabled()
		return not runningModules[moduleName]
	end

	options.args.track.args[moduleName] = {
		type = "toggle",
		name = displayName,
		desc = L["Toggle whether to track %q."]:format(displayName),
		order = order,
	}

	options.args.colors.args[moduleName] = {
		type = "color",
		name = displayName,
		desc = L["What color to use for the %q data type."]:format(displayName),
		order = 100,
		hidden = disabled,
		order = order,
	}

	options.args.fubar.args[moduleName] = {
		type = "toggle",
		name = displayName,
		desc = L["Toggle whether to display data from %q on the FuBar panel."]:format(displayName),
		hidden = disabled,
		order = order,
	}

	options.args.tooltip.args[moduleName] = {
		type = "toggle",
		name = displayName,
		desc = L["Toggle whether to display data from %q in the tooltip."]:format(displayName),
		hidden = disabled,
		order = order,
	}

	options.args.report.args[moduleName] = {
		type = "group",
		name = displayName,
		desc = L["Send %s data."]:format(displayName),
		guiDesc = L["Send %s data.\nClick to send to most recent destination."]:format(displayName),
		handler = mod,
		hidden = disabled,
		order = order,
		args = reportToMenu,
		pass = true,
		onClick = "onClickReporting",
		set = "setReporting",
		get = "getReporting",
	}


	if mod.dataFormatter then
		if type(mod.dataFormatter) == "function" then
			dataFormatter[mod.name] = mod.dataFormatter
		elseif type(mod.dataFormatter) == "string" then
			dataFormatter[mod.name] = function(v) return mod.dataFormatter:format(v) end
		end
	end

	if type(mod.OnModuleRegister) == "function" then
		mod:OnModuleRegister()
	end
end

function Violation:GetModuleVersion()
	return moduleVersion
end

function Violation:ModuleIncompatible(mod)
	self:ToggleModuleActive(mod, false)
	self:Print(L["%s has been disabled, since it is not compatible with the latest module API."]:format(mod.name))
end

function Violation:ToggleModule(mod, val)
	db.enabledModules[mod] = val
	Violation:ToggleModuleActive(mod, val)
	Violation:UpdateDisplay()
end

do
	local function sorter(a, b)
		local moduleA = runningModules[a]
		local moduleB = runningModules[b]
		local diff = tonumber(moduleA.violationOrdering) - tonumber(moduleB.violationOrdering)
		return (diff == 0) and (moduleA:GetDisplayName() < moduleB:GetDisplayName()) or (diff < 0)
	end

	function Violation:UpdateRunningModuleOrdering()
		runningModuleOrdering = del(runningModuleOrdering)
		runningModuleOrdering = new()
		for name, module in pairs(runningModules) do
			table_insert(runningModuleOrdering, module.name)
		end
		table.sort(runningModuleOrdering, sorter)
	end
end

function Violation:IsModuleRunning(name)
	return runningModules[name] and true or false
end

function Violation:IterateRunningModules()
	return ipairs(runningModuleOrdering)
end


-----------------------------------------------------------------------
-- Pet merging stuff
-----------------------------------------------------------------------

function Violation:OnRosterUpdate(first)

	-- are we grouped, we always keep track of this regardless of the showGrouped setting
	if (grouped or first) and (GetNumPartyMembers() == 0 and GetNumRaidMembers() == 0) then
		grouped = nil
		if db.showGrouped then self:ToggleWindows(false) end
	elseif (not grouped or first) and (GetNumPartyMembers() > 0 or GetNumRaidMembers() > 0) then
		grouped = true
		if db.showGrouped then self:ToggleWindows(true) end
	end

	-- Update characterPets for named pets (warlocks' and hunters' ones)

	local seen = new()
	local dirty = false

	-- Scan roster
	for unit in roster:IterateRoster() do
		-- Gotya
		seen[unit.name] = true
		characterDataSource[unit.name] = unit.name

		-- Record hunter and warlock pets
		if unit.unitid and (unit.class == "WARLOCK" or unit.class == "HUNTER") then
			-- Check for current pet
			local petId = roster:GetPetFromOwner(unit.unitid)
			if petId then
				local pet = roster:GetUnitObjectFromUnit(petId)
				if pet then
					if not characterPets[unit.name] then
						characterPets[unit.name] = new()
					end
					if not characterPets[unit.name][pet.name] then
						-- This is very ugly, but needed...
						characterPets[unit.name][pet.name] = true
						table_insert(characterPets[unit.name], pet.name)
						dirty = true
					end
				end
			end
		-- Ignore pets of other classes
		elseif characterPets[name] then
			characterPets[name] = del(characterPets[name])
			dirty = true
		end
	end

	-- Remove old data
	for name in pairs(characterPets) do
		if not seen[name] then
			characterPets[name] = del(characterPets[name])
			characterDataSource[name] = nil
			dirty = true
		end
	end

	-- Link pets to master
	for master, pets in pairs(characterPets) do
		for i,pet in ipairs(pets) do
			characterDataSource[pet] = master
		end
	end

	seen = del(seen)

	if dirty then
		self:UpdateDisplay()
		self:UpdateWindows()
	end
end

function Violation:GetCharacterPet(charName)
	if characterPets[charName] then
		for k, v in pairs(characterPets[charName]) do
			return v
		end
	end
	return
end

function Violation:GetPetMaster(petName)
	return characterDataSource[petName]
end

function Violation:IsAnonymousPetEvent(info)
	return info.event:match("PET")
			and info.eventType == "Damage"
			and info.sourceName
			and characterDataSource[info.sourceName] ~= playerName
end

function Violation:GetModuleData(modName, charName)
	local mod = runningModules[modName]
	if not mod then
		return
	end

	if db.mergePets and mod.canMergeData then
		-- Specific player data
		if charName then
			-- Merge if necessary
			if characterPets[charName] then
				return mod:GetMergedData(charName, unpack(characterPets[charName]))
			end
		-- All module data
		else
			-- Clean previous data table
			for k, v in pairs(moduleData) do
				moduleData[k] = nil
			end

			-- Clone module data
			for k,v in pairs(mod:GetData()) do
				moduleData[k] = v
			end

			-- Merge data where needed
			for main, others in pairs(characterPets) do
				-- Delete pets from moduleData
				for k, v in pairs(others) do
					moduleData[v] = nil
				end
				moduleData[main] = mod:GetMergedData(main, unpack(others))
			end
			return moduleData
		end
	end
	-- Fallback to old, plain data
	return mod:GetData(charName)
end

-----------------------------------------------------------------------
--
-----------------------------------------------------------------------

local function Sorter(a, b) return a[2] > b[2] end

-- this dataInfo table is reused throughout violation for the current information being displayed.
local dataInfo = {}
function Violation:PublishModuleData(modName, AddTabletRowCallback, count, offset, window)
	local mod = runningModules[modName]
	if not mod then
		return
	end
	local data = self:GetModuleData(modName)
	assert(type(data) == "table", "Some module provided us with nil data.")

	local tbl = new()
	dataInfo.total = 0
	dataInfo.maxValue = 0
	for k, v in pairs(data) do
		if type(v) == "number" and v > 0 then
			dataInfo.total = dataInfo.total + v
			dataInfo.maxValue = math.max(dataInfo.maxValue, v)
			table_insert(tbl, new(k, v))
		end
	end

	local tblCount = #tbl

	count = math.min(count or tblCount, tblCount)
	offset = math.max(math.min(offset or 0, tblCount-count), 0)

	local n = 0
	if tblCount > 0 then
		local percentage = nil

		if tblCount > 1 then
			table.sort(tbl, Sorter)

			if not mod.hidePercentage then
				percentage = new()
				for i, v in ipairs(tbl) do
					table_insert(percentage, (v[2] * 100) / dataInfo.total)
				end
			end
		end

		dataInfo.hasMergedData = db.mergePets and mod.canMergeData and next(characterPets) and true
		dataInfo.rowCount = tblCount
		dataInfo.module = mod
		dataInfo.formatter = dataFormatter[modName] or tostring
		dataInfo.hasPercentage = percentage

		for i = 1, count do
			local rank = offset + i
			local name, value = unpack(tbl[rank])
			if not name then
				break
			end
			if window then
				AddTabletRowCallback(window, i, rank, name, value, percentage and percentage[rank])
			else
				AddTabletRowCallback(i, rank, name, value, percentage and percentage[rank])
			end
			n = n + 1
		end

		percentage = del(percentage)
	end

	tbl = del(tbl)
	
	return n, offset
end

-----------------------------------------------------------------------
-- Channel output
-----------------------------------------------------------------------

function Violation:PrintModuleDataToChannel(modName, channel, target)

	-- This should stay here (depends on channel)
	local function doSend(msg, ...)
		if select("#", ...) > 0 then
			msg = msg:format(...)
		end
		ChatThrottleLib:SendChatMessage("BULK", nil, msg, channel, nil, target)
	end
	dataInfo.header = nil
	
	self:PublishModuleData(
		modName,
		function(i, rank, name, value, percentage)
			if not dataInfo.header then
				local name = dataInfo.module:GetDisplayName()
				if dataInfo.hasMergedData then
					doSend(L["==== %s (pets merged) ===="], name)
				else
					doSend(L["==== %s ===="], name)
				end
				dataInfo.header = true
			end
			local value = dataInfo.formatter(value)
			if percentage then
				doSend("%d. %s - %s "..percentString, rank, name, value, percentage)
			else
				doSend("%d. %s - %s", rank, name, value)
			end
		end,
		db.maxPrintLines
	)
end

-----------------------------------------------------------------------
-- UI
-----------------------------------------------------------------------

local classOrders = {
	WARRIOR = 100,
	ROGUE = 200,
	SHAMAN = 300,
	HUNTER = 400,
	MAGE = 500,
	WARLOCK = 600,
	DRUID = 700,
	PRIEST = 800,
	PALADIN = 900,
}

local uniquePartyUpdate=12345

function Violation:UpdatePartyMenu()
	uniquePartyUpdate = uniquePartyUpdate + 1
	local whisperTargets = reportToMenu.whisper.args

	for unit in roster:IterateRoster() do
		if unit and unit.class and unit.class ~= "PET" then
			local name = unit.name
			if not whisperTargets[name] then
				local option = new()
				option.unique=uniquePartyUpdate
				option.type = "toggle"
				option.name = coloredNames[name]
				option.desc = L["Send reports to %s."]:format(name)
				option.passValue = newkeyed("method","WHISPER","target",name)
				option.order = classOrders[unit.class]
				whisperTargets[name] = option
			else
				whisperTargets[name].unique = uniquePartyUpdate
			end
		end
	end

	for k,o in pairs(whisperTargets) do
		if o.unique and o.unique~=uniquePartyUpdate then
			whisperTargets[k] = del(whisperTargets[k])
		end
	end

end

function Violation:OnClick()
	if IsAltKeyDown() then
		self:ResetGroup()
	elseif IsShiftKeyDown() then
		self:Reset()
	else
		self:ToggleWindows(not db.showWindows)
	end
end

function Violation:UpdateDisplayIfDirty()
	local dirty = nil
	local dirtpile = new()
	for k, v in pairs(db.tooltipModules) do
		if v and runningModules[k] and runningModules[k]:IsDirty() then
			table_insert(dirtpile, k)
			dirty = true
		end
	end
	if dirty then
		self:UpdateDisplay()
	end

	self:UpdateDirtyWindows()

	for k, v in ipairs(dirtpile) do
		runningModules[v]:ResetDirty()
	end
	dirtpile = del(dirtpile)
end

local tooltipColors = {
	"child_textR", 1,
	"child_textG", 1,
	"child_textB", 0,
	"child_text2R", 1,
	"child_text2G", 1,
	"child_text2B", 1,
	"child_text3R", 1,
	"child_text3G", 1,
	"child_text3B", 1,
	"child_text4R", 0.65,
	"child_text4G", 0.65,
	"child_text4B", 0.65
}

-- row adding callback
local function AddTabletRowCallback(i, rank, name, value, percentage)
	-- Create header on the fly
	if not dataInfo.cat then
		local displayName = dataInfo.module:GetDisplayName()
		local tblCount = dataInfo.rowCount
		dataInfo.cat = tablet:AddCategory(
			"columns", 2 + (tblCount > 1 and 1 or 0) + (dataInfo.hasPercentage and 1 or 0),
			"text", tblCount > 1 and L["#"] or displayName,
			"text2", tblCount > 1 and L["Player"] or "",
			"text3", displayName,
			"child_justify", "LEFT",
			"child_justify2", tblCount > 1 and "LEFT" or "RIGHT",
			"child_justify3", "RIGHT",
			"child_justify4", "RIGHT",
			unpack(tooltipColors)
		)
	end

	-- Add row
	if dataInfo.rowCount > 1 then
		dataInfo.cat:AddLine(
			"text", tostring(rank) .. ".",
			"text2", coloredNames[name],
			"text3", dataInfo.formatter(value or 0),
			"text4", percentage and percentString:format(percentage)
		)
	else
		dataInfo.cat:AddLine(
			"text", coloredNames[name],
			"text2", dataInfo.formatter(value or 0)
		)
	end
end

function Violation:OnTooltipUpdate()
	local no = true
	for k, v in pairs(db.tooltipModules) do
		if v and runningModules[k] then
			no = nil
			dataInfo.cat = nil
			self:PublishModuleData(k, AddTabletRowCallback, 10, 0)
		end
	end
	if no then
		tablet:AddCategory("columns", 1):AddLine("text", L["No modules running, please enable one or more from the configuration menu."])
	end
	if updateWarned then
		tablet:SetHint(L["|cffeda55fClick|r to toggle the bar display. |cffeda55fAlt-Click|r to issue a group reset. |cffeda55fShift-Click|r to reset."] .."\n|cffffff00"..L["Violation v%s is available. Please consider updating."]:format(tostring(latestRevisionSeen)).."|r")
	else
		tablet:SetHint(L["|cffeda55fClick|r to toggle the bar display. |cffeda55fAlt-Click|r to issue a group reset. |cffeda55fShift-Click|r to reset."])
	end
end

function Violation:OnTextUpdate()
	local hasText = nil
	for k, v in pairs(db.textModules) do
		if v and runningModules[k] then
			local data = self:GetModuleData(k, playerName)
			if type(data) == "number" and data > 0 then
				local shortName = runningModules[k]:GetShortDisplayName()
				local color = db.colors[k]
				data = (dataFormatter[k] or tostring)(data)
				local text = Colorize(string.format("%s: %s", shortName, data), color.r, color.g, color.b)
				if not hasText then hasText = new(text)
				else table_insert(hasText, text) end
			end
		end
	end

	if not hasText then
		self:SetText(self.name)
	else
		self:SetText(table.concat(hasText, " "))
		hasText = del(hasText)
	end
end

-----------------------------------------------------------------------
-- Data broadcasting and recieving
-----------------------------------------------------------------------

function Violation:BroadcastAll()
	self:Broadcast()
	local petName = self:GetCharacterPet(playerName)
	if petName then
		self:BroadcastFor(petName)
	end
end

function Violation:BroadcastFor(charName)
	local tbl, data = nil, nil
	local i = 0
	for name, mod in pairs(runningModules) do
		if mod.canBroadcast and mod:IsBroadcastDirty(charName) then
			if not tbl then tbl = new() end
			i = i + 1
			table_insert(tbl, name)
			table_insert(tbl, mod:GetBroadcastData(charName))
			if i == 9 then -- we send max 18 pairs + "BF" to stay under the limit of 20 args in acecomm
				self:SendCommMessage("GROUP", "BF", charName, unpack(tbl))
				tbl = del(tbl)
				i = 0
			end
		end
	end
	if tbl then
		self:SendCommMessage("GROUP", "BF", charName, unpack(tbl))
	end
	tbl = del(tbl)
end

function Violation:Broadcast()
	local tbl, data = nil, nil
	local i = 0
	for name, mod in pairs(runningModules) do
		if mod.canBroadcast and mod:IsBroadcastDirty() then
			if not tbl then tbl = new() end
			i = i + 1
			table_insert(tbl, name)
			table_insert(tbl, mod:GetBroadcastData())
			if i == 9 then -- we send max 18 pairs + "B" to stay under the limit of 20 args in acecomm
				self:SendCommMessage("GROUP", "B", unpack(tbl))
				tbl = del(tbl)
				i = 0
			end
		end
	end
	if tbl then
		self:SendCommMessage("GROUP", "B", unpack(tbl))
	end
	tbl = del(tbl)
end

function Violation:ResetGroup()
	if not IsRaidLeader() and not IsRaidOfficer() and not IsPartyLeader() then
		self:Print(L["Only the group leader or an officer can do that."])
		return
	end

	self:SendCommMessage("GROUP", "LeaderReset")
	self:Reset()
end

function Violation:UpdateWarning(version, revision)
	if revision > self.revision and not updateWarned then
		updateWarned = true
		self:Print(L["Violation v%s is available. Please consider updating."], version)
	end
end

function Violation:OnPrivateComm(prefix, sender, distribution, command, ...)
	if distribution ~= "WHISPER" or type(command) ~= "string" or type(sender) ~= "string" then return end
	if command == "ViolationPong" then
		if type(violationUsers) ~= "table" then
			violationUsers = new()
		end
		if violationUsers[sender] then
			violationUsers[sender] = del(violationUsers[sender])
		end
		violationUsers[sender] = new()
		local version, revision = select(1,...)
		revision = tonumber(revision)
		violationUsers[sender].version = version
		if revision > (latestRevisionSeen or 0) then
			latestRevisionSeen = revision
			self:ScheduleEvent("Violation_UpdateWarning", self.UpdateWarning, 5, self, version, revision)
		end
		violationUsers[sender].revision = revision
		violationUsers[sender].globalLevel = VIOLATION_LEVEL
		for i = 3, select("#", ...) do
			local x = (select(i, ...))
			if x then
					violationUsers[sender][x] = VIOLATION_LEVEL
			end
		end
	end
end

function Violation:OnComm(prefix, sender, distribution, command, ...)
	if distribution ~= "GROUP" or type(command) ~= "string" then return end

	if command == "B" then
		for i = 1, select("#", ...), 2 do
			local name, value = select(i, ...)
			if value and tonumber(value) and tonumber(value) > 0 then
				self:SetUserViolation(sender, name, VIOLATION_LEVEL)
				if runningModules[name] then
					runningModules[name]:OnRemoteData(sender, value)
				end
			end
		end
	elseif command == "BF" then
		local charName = select(1, ...)
		for i = 2, select("#", ...), 2 do
			local name, value = select(i, ...)
			if value and tonumber(value) and tonumber(value) > 0 then
				self:SetUserViolation(charName, name, VIOLATION_PROXY_LEVEL)
				if runningModules[name] then
					runningModules[name]:OnRemoteData(charName, value)
				end
			end
		end
	elseif command == "LeaderReset" and db.leaderReset then
		local obj = roster:GetUnitObjectFromName(sender)
		if obj and obj.rank > 0 then
			self:Print(L["Resetting data as requested by %s."]:format(sender))
			self:Reset()
		end
	elseif command == "ViolationPing" then
		local tbl = new()
		for k in pairs(runningModules) do
			table_insert(tbl, k)
		end
		if #tbl > 0 then
			self:SendCommMessage("WHISPER", sender, "ViolationPong", self.version, self.revision, unpack(tbl))
		end
		tbl = del(tbl)
	end
end

function Violation:UserViolation(user, mod, level)
	if type(violationUsers) ~= "table" or type(violationUsers[user]) ~= "table" then return end
	if not level then level = COMBATLOG_LEVEL end
	if mod and violationUsers[user][mod] and violationUsers[user][mod] > level then
		return true
	elseif not mod and violationUsers[user] then
		return true
	end
	return false
end

function Violation:SetUserViolation(user, mod, level)
	if not user or not mod or level == nil then return end
	if type(violationUsers) ~= "table" then
		violationUsers = new()
	end
	if type(violationUsers[user]) ~= "table" then
		violationUsers[user] = new()
	end
	if level > (violationUsers[user].globalLevel or 0) then
		violationUsers[user].globalLevel = level
	end
	violationUsers[user][mod] = level
end

-- Check if :
-- * both target and source are in party,
-- * source may provide data for target,
-- * source has sufficient level to provide data for target, according to module and level
-- source defaults to local player,
-- level defaults to VIOLATION_LEVEL for local player and to COMBATLOG_LEVEL for others
function Violation:AcceptDataFromSource(destination, module, source, level)

	-- We're only interested of destination inside the party
	local regularSource = characterDataSource[destination]
	if not regularSource then
		return false
	end

	-- Source is local player
	if not source or source == playerName then

		-- Player is the regular source of target: always accept
		if regularSource == playerName then
			return true
		end
		-- Player isn't the regular source, do the check anyway

	-- Source isn't the regular source and isn't local player, drop data
	elseif regularSource ~= source then
		return false
	end

	-- Check if regular source has the sufficient level
	return not self:UserViolation(regularSource, module, level)
end

function Violation:GetUserDataSourceDesc(user, module)
	if user == playerName or characterDataSource[user] == playerName then
		return VIOLATION_LEVEL, "Violation v"..self.version
	end
	local data = violationUsers and violationUsers[user]
	local level = data and data[module or 'globalLevel']
	if level == VIOLATION_LEVEL then
		return VIOLATION_LEVEL, "Violation v"..(data.version or "?")
	elseif level == VIOLATION_PROXY_LEVEL then
		return VIOLATION_PROXY_LEVEL, "Violation Proxied"
	elseif level == SWS_LEVEL then
		return SWS_LEVEL, "SWStats"
	elseif level == DAMAGEMETERS_LEVEL then
		return DAMAGEMETERS_LEVEL, "DamageMeters"
	else
		return COMBATLOG_LEVEL, "Combat Log"
	end
end

function Violation:ReportUsers()
	if violationUsers and next(violationUsers) then
		for unit in roster:IterateRoster() do
			local user = unit.name
			if user ~= playerName then
				local level, source = self:GetUserDataSourceDesc(user)
				local msg = user..": "..source
				if level == VIOLATION_LEVEL then
					local tbl = new()
					for mod,v in pairs(violationUsers[user]) do
						if type(mod) == "string" and v == VIOLATION_LEVEL and mod ~= "globalLevel" and mod ~= 'revision' and mod ~= 'version' then
							table_insert(tbl, mod)
						end
					end
					if #tbl then
						msg = L["%s, modules: %s"]:format(msg, table.concat(tbl, ", "))
					end
					tbl = del(tbl)
				end
				self:Print(msg)
			end
		end
	else
		self:Print(L["No syncing information"])
	end
end

-----------------------------------------------------------------------
-- Spying support
-----------------------------------------------------------------------

function Violation:UpdateSpying()

	-- (Un)register events
	if (db.spying.damageMeters or db.spying.swStats) then
		if not self:IsEventRegistered("CHAT_MSG_ADDON") then
			self:RegisterEvent("CHAT_MSG_ADDON", "OnRawComm")
		end
	elseif self:IsEventRegistered("CHAT_MSG_ADDON") then
		self:UnregisterEvent("CHAT_MSG_ADDON")
	end

	-- Reset UserViolation data if needed
	if violationUsers then
		for user,data in pairs(violationUsers) do
			local maxLevel = COMBATLOG_LEVEL
			for mod,level in pairs(data) do
				if mod ~= "version" and mod ~= "revision" and type(level) == "number" then
					if (level == SWS_LEVEL and not db.spying.swStats)
						or (level == DAMAGEMETERS_LEVEL and not db.spying.damageMeters) then
						violationUsers[user][mod] = nil
					elseif level > maxLevel then
						maxLevel = level
					end
				end
			end
			if maxLevel == COMBATLOG_LEVEL then
				violationUsers[user] = del(violationUsers[user])
			else
				data.globalLevel = maxLevel
			end
		end
	end
end

Violation.commSpies = {}

function Violation:OnRawComm(prefix, message, distribution, sender)
	local spy = self.commSpies[prefix]

	if not spy or (distribution ~= "PARTY" and distribution ~= "RAID" and distribution ~= "BATTLEGROUND") then
		return
	end

	if sender == playerName or not roster:GetUnitObjectFromName(sender) then
		return
	end
	return spy(self, message, distribution, sender)
end

-----------------------------------------------------------------------
-- DamageMeters supports
-----------------------------------------------------------------------

-- Damage meter sends a triplet (total/hit/crit) for each of these data
local damageMeterInfo = {
	-- Protocol version 12
	[12] = { "Damage", "Healing", "DamageTaken", "HealingTaken", "Decurse", "Overhealing" },
}

-- Hold current DM session values
local damageMeterSession

function Violation.commSpies:DamageMeters(message, distribution, sender)
	if not db.spying.damageMeters then
		return
	end

	-- Parse message
	local command, version, arguments = select(3, message:find("(%w+)_(%d+)_ (.*)$"))

	-- Check version
	version = version and tonumber(version)
	if not version or not damageMeterInfo[version] then
		return
	end

	if command == "REQ_CLEAR" then
		-- Clear request
		local sessionId = select(3, arguments:find("^(.+ %d+)$"))
		if sessionId then
			damageMeterSession = del(damageMeterSession)
		end

	elseif command == "REQ_SYNC" then
		-- Sync request, check session

		local sessionId = select(3, arguments:find("^<(.+ %d+)>$"))
		if sessionId and damageMeterSession and damageMeterSession.sessionId ~= sessionId then
			-- Sessions mismatch, drop all data
			damageMeterSession = del(damageMeterSession)
		end

	elseif command == "SYNC" then
		-- Sync data
		local name, data = select(3, arguments:find("^(%S+) (.+)$"))

		-- Data parsing
		local dmi = damageMeterInfo[version]
		local index = 1
		for value in data:gmatch("(%d+) %d+ %d+") do
			if index > #dmi then
				break
			end
			value = tonumber(value)
			if value > 0 then
				local modName = dmi[index]

				if self:AcceptDataFromSource(name, modName, sender, DAMAGEMETERS_LEVEL) then
					self:SetUserViolation(sender, modName, DAMAGEMETERS_LEVEL)

					-- Vivify session data
					if not damageMeterSession then
						damageMeterSession = new()
					end
					if not damageMeterSession[modName] then
						damageMeterSession[modName] = new()
					end

					-- Only feed Violation if we are in a session
					if damageMeterSession.sessionId and runningModules[modName] then
						local oldValue = (damageMeterSession[modName][name] or 0)
						runningModules[modName]:OnRemoteData(name, value-oldValue)
					end

					-- Record value for future use
					damageMeterSession[modName][name] = value
				end

			end
			index = index + 1
		end

		-- Create a fake session if needed (until someone send us an real session id)
		if damageMeterSession and not damageMeterSession.sessionId then
			damageMeterSession.sessionId = "FAKE-"..tostring(damageMeterSession)
		end
	end
end


-- Hold current SWS session values
local SWSSession = {}
local SWSSyncMessage = "(%d+)"..string.char(31).."(.+)"
local SWSSplitMessage = "(.+)"..string.char(30).."(.+)"..string.char(30).."(.+)"
local SWSSplitSync = "[^"..string.char(29).."]+"

local SWSmodulesOUT = {
	"Damage done",
	"Healing done",
	false, -- "Deaths",
	"Decursing done",
	false, --"TotalDmgTicks",
	"Overhealing done",
	false, -- "HealInFight",
	false, -- "OverHealInFight",
	false, -- "DOTTicks",
	false, -- "DOTValue", -- SWS records dots separate from damage, we handle this in the main sync handler
	false, -- "TotalFightTime",
}

local SWSmodulesIN = {
	"Damage taken",
	"Healing taken",
	false, -- "Deaths",
	false, -- "Decurse",
	false, --"TotalDmgTicks",
	false, -- "OverHealing",
	false, -- "HealInFight",
	false, -- "OverHealInFight",
	false, -- "DOTTicks",
	false, -- "DOTValue", -- SWS records dots separate from damage, we handle this in the main sync handler
	false, -- "TotalFightTime",
}

-- If the module is currently running has already been asserted when calling this function
local function AddSWSData(sender, name, modName, value)
	if Violation:AcceptDataFromSource(name, modName, sender, SWS_LEVEL) then
		Violation:SetUserViolation(sender, modName, SWS_LEVEL)
		-- self:Print("Received "..tostring(value).." SWS from "..sender.." for "..modName)
		if not SWSSession[modName] then
			SWSSession[modName] = {}
		end
		local oldValue = SWSSession[modName][name]
		-- this way we ignore the first received data from someone and update afterwards
		-- this mimics Violation mechanics better
		if oldValue and oldValue < value then
			runningModules[modName]:OnRemoteData(name, value-oldValue)
		end
		SWSSession[modName][name] = value
	end
end

local function HandleSWSMessage(sender, msg)
	local type, name, data = select(3, msg:find(SWSSplitMessage))
	if not name or (name ~= sender and Violation:GetPetMaster(name) ~= sender) then
		return
	end
	if type and type == "BUDD" or type == "BUDR" then
		local i = 0
		local numbers = new()
		for tmp in data:gmatch("%d+") do
			i = i + 1
			numbers[i] = tonumber(tmp)
		end
		numbers[1] = numbers[1] + numbers[10]			-- damage + dotdamage
		numbers[6] = numbers[6] + numbers[8]			-- overhealing + overhealinginfight
		numbers[2] = numbers[2] + numbers[7] - numbers[6]	-- healing + healinginfight - overhealing
		local mod
		for k, value in ipairs(numbers) do
			mod = type == "BUDD" and SWSmodulesOUT[k] or SWSmodulesIN[k]
			if mod and runningModules[mod] then
				AddSWSData(sender, name, mod, value)
			end
		end
		numbers = del(numbers)
	end
end

-- General SWS message handler
function Violation.commSpies:SWSudX21(message, distribution, sender)
	if not db.spying.swStats then
		return
	end
	local session, data = select(3, message:find(SWSSyncMessage))
	if not session or not data then
		return
	end
	for msg in data:gmatch(SWSSplitSync) do
		HandleSWSMessage(sender, msg)
	end
end

-- SWS Handshake handler, we forward this to the main handler, since we handle everything there anyway
function Violation.commSpies:SWSHSX21(message, distribution, sender)
	local spy = self.commSpies["SWSudX21"]
	spy(self, message, distribution, sender)
end



----------------------------
-- GUI STUFF
----------------------------

---------------------------
-- Locals
---------------------------

local cachedBars = {}
local windowCache = {}

Violation.windows = {}

----------------------------
-- Violation Window Menu Template and helpers
----------------------------

local function copyTable(source)
	local result = {}
	for k, v in pairs(source) do
		if type(v) == "table" then
			result[k] = copyTable(v)
		else
			result[k] = v
		end
	end
	return result
end


local windowDefaults = {
	width = 200,
	height = 200,
	nrBars = 10,
	scale = 1,
	show = true,
	module = "Damage",
	locked = false,
	useModuleColor = true,
	color = { r = 1, g = 0, b = 0 },
	barTexture = "Blizzard",
	useDefaultTexture = true,
	useSimpleBars = false,
	posX = false,
	posY = false,
}

local windowOptionsTemplate = {
	report = {
		type = "group",
		name = L["Report"],
		desc = L["Report your current data to a channel."],
		guiDesc = L["Report your current data to a channel.\nClick to send to most recent destination."],
		hidden = disabled,
		args = reportToMenu,
		pass = true,
		onClick = "onClickReporting",
		set = "setReporting",
		get = "getReporting",
		order = 100,
	},
	show = {
		type = "toggle",
		name = L["Show"],
		desc = L["Show Window"],
		get = "windowGet",
		set = "windowSet",
		passValue = "show",
		order = 105,
	},
	lock = {
		type = "toggle",
		name = L["Lock"],
		desc = L["Lock Window"],
		get = "windowGet",
		set = "windowSet",
		passValue = "locked",
		order = 110,
	},
	layoutoptionsspacer = {
		type = "header",
		text = " ",
		order = 115,
	},
	modulecolor = {
		type = "toggle",
		name = L["Use Module Color"],
		desc = L["Use Module Color for the window header"],
		get = "windowGet",
		set = "windowSetSoft",
		passValue = "useModuleColor",
		order = 120,
	},
	color = {
		type = "color",
		name = L["Color"],
		desc = L["Window Color"],
		get = "windowGetColor",
		set = "windowSetColor",
		disabled = "windowUsesModuleColor",
		passValue = "color",
		order = 121,
	},
	defaulttexture = {
		type = "toggle",
		name = L["Use default texture"],
		desc = L["Use default bar texture"],
		get = "windowGet",
		set = "windowSetSoft",
		passValue = "useDefaultTexture",
		order = 130,
	},
	texture = {
		type = "text",
		name = L["Texture"],
		desc = L["Bar Texture"],
		get = "windowGet",
		set = "windowSetSoft",
		passValue = "barTexture",
		validate = SM:List("statusbar"),
		disabled = "windowUsesDefaultTexture",
		order = 131,
	},
	simple = {
		type = "toggle",
		name = L["Show simple bar values"],
		desc = L["Show simple bar values"],
		get = "windowGet",
		set = "windowSetSoft",
		passValue = "useSimpleBars",
		order = 132,
	},
	scale = {
		type = "range",
		name = L["Scale"],
		desc = L["Set the window scale"],
		min = 0.4, max=2, step=0.02, bigStep=0.1, isPercent=true,
		get = "windowGet",
		set = "windowSetSoft",
		passValue = "scale",
		order = 140,
	},
	resetspacer = {
		type = "header",
		name = " ",
		order = 190,
	},
	reset = {
		type = "execute",
		name = L["Reset module"],
		desc = L["Reset window module data"],
		func = "windowReset",
		passValue = "reset",
		order = 191,
	},
	delspacer = {
		type = "header",
		name = " ",
		order = 200,
	},
	delete = {
		type = "execute",
		name = L["Delete"],
		desc = L["Delete Window"],
		func = "windowDelete",
		passValue = "delete",
		order = 201,
	},
}

---------------------------
-- Initialization from Violation
---------------------------

function Violation:SetupWindows()
	for i = 1, db.windows.count do
		if not self.windows[i] then
			if not db.windows[i] then
				db.windows[i] = copyTable(windowDefaults)
			end
			self.windows[i] = self:CreateWindow(db.windows[i])
			-- we call setup here because at this point getID() is available.
			self.windows[i]:Setup()
		end
	end
end

function Violation:CreateWindow(options)
	local w
	if #windowCache > 0 then
		w = table.remove(windowCache)
	else
		w = self.VioWindow:new()
	end
	w.options = options
	return w
end

function Violation:DeleteWindow(window)
	window:Delete()
	local wdb = db.windows
	local oldindex = window:GetID()
	for i = oldindex, wdb.count - 1 do
		wdb[i] = wdb[i+1]
		self.windows[i] = self.windows[i+1]
	end
	table_insert(windowCache, window)
	wdb[wdb.count] = nil
	self.windows[wdb.count] = nil
	options.args.windows.args[wdb.count] = nil
	wdb.count = wdb.count - 1
	for i = 1, wdb.count do
		self.windows[i]:SetupMenuOptions()
	end
	return nil
end

function Violation:UpdateWindows(sameamount)
	for k, w in ipairs(self.windows) do
		w:Update(sameamount)
	end
end

function Violation:UpdateDirtyWindows()
	local dirtpile = new()
	for k, w in ipairs(self.windows) do
		local modName = w:UpdateIfDirty()
		if modName then
			table_insert(dirtpile, modName)
		end
	end
	for k, modName in ipairs(dirtpile) do
		local mod = self:GetModule(modName)
		mod:ResetDirty()
	end
	dirtpile = del(dirtpile)
end

function Violation:ToggleWindows(show)
	local shown=0
	for k, w in ipairs(self.windows) do
		if w.options.show then
			shown=shown+1
		end
	end
	if shown<1 then
		Violation:Print(L["All windows are currently configured to be not Shown. You will need to show some windows manually before you can use this shortcut to temporarily hide and show windows."])
		db.showWindows = true
		return
	end

	db.showWindows = show
	if show then
		self:OpenWindows()
	else
		self:CloseWindows()
	end
end

function Violation:CloseWindows()
	for k, w in ipairs(self.windows) do
		w:Close()
	end
end

function Violation:OpenWindows()
	for k, w in ipairs(self.windows) do
		if w.options.show then
			w:Open()
		end
	end
end

----------------------------
-- Bars Local Helpers
----------------------------

local OnBarEnter, OnBarLeave -- early declaration

local function NewBar(texture)
	local f
	if #cachedBars > 0 then
		f = table.remove(cachedBars)
	else
		f = CreateFrame("Button", nil, UIParent)
	end
	f:Hide()
	f:SetParent(UIParent)

	-- f:SetWidth(150)
	f:SetHeight(15)

	if not f.statusbar then
		f.statusbar = CreateFrame("StatusBar", nil, f)
	end
	f.statusbar:ClearAllPoints()
	f.statusbar:SetAllPoints(f)
	f.statusbar:SetStatusBarTexture(SM:Fetch("statusbar", texture))
	f.statusbar:SetStatusBarColor(.5, .5, .5, 1)
	f.statusbar:SetMinMaxValues(0,100)
	f.statusbar:SetValue(0)

	if not f.lefttext then
		f.lefttext = f.statusbar:CreateFontString(nil, "OVERLAY")
	end
	f.lefttext:SetFontObject(GameFontHighlightSmall)
	f.lefttext:SetPoint("LEFT", f.statusbar, "LEFT", 0, 0)
	f.lefttext:SetJustifyH("LEFT")
	f.lefttext:SetText("")
	f.lefttext:SetTextColor(1,1,1,1)

	if not f.righttext then
		f.righttext = f.statusbar:CreateFontString(nil, "OVERLAY")
	end
	f.righttext:SetFontObject(GameFontHighlightSmall)
	f.righttext:SetPoint("RIGHT", f.statusbar, "RIGHT", 0, 0)
	f.righttext:SetJustifyH("RIGHT")
	f.righttext:SetText("")
	f.righttext:SetTextColor(1, 1, 1, 1)

	if not f:GetScript("OnEnter") then
		f:SetScript("OnEnter", OnBarEnter)
	end
	if not f:GetScript("OnLeave") then
		f:SetScript("OnLeave", OnBarLeave)
	end

	return f
end

local function FreeBar(f)
	f:Hide()
	f:SetParent(UIParent)
	f:ClearAllPoints()
	table_insert(cachedBars, f)
	return nil
end

----------------------------
-- Vio Window Local Helpers
----------------------------

local function OnWindowMouseUp(this, button)
	if button == "RightButton" then
		this.window:OpenMenu()
	end
end

local function OnWindowMouseWheel(this, dir)
	this.window:OnOffset(dir > 0)
end

local function OnWindowClose(this)
	this:GetParent().window.options.show = false
	this:GetParent().window:Close()
end

local function OnWindowPrevModule(this)
	this:GetParent().window:SwitchView(false)
end

local function OnWindowNextModule(this)
	this:GetParent().window:SwitchView(true)
end

local function OnWindowDragStop(this)
	this:StopMovingOrSizing()
	local s = this:GetEffectiveScale()
	this.window.options.posX = this:GetLeft() * s
	this.window.options.posY = this:GetTop() * s
end

local function OnWindowSizeChanged(this)
	this.window:OnSized()
end

local function OnDragHandleShow(this)
	this.draghandle:Show()
end

local function OnDragHandleHide(this)
	this.draghandle:Hide()
end

local function ScheduleDragHandleHide(this)
	if this.windowframe then
		this = this.windowframe
	end
	if this.hideEvent then
		Violation:CancelScheduledEvent(this.hideEvent)
		this.hideEvent = nil
	end
	this.hideEvent = "Violation-" .. math.random()
	Violation:ScheduleEvent(this.hideEvent, OnDragHandleHide, 1, this)
end

local function OnDragHandleMouseDown(this)
	this.windowframe.isResizing = true
	this.windowframe:StartSizing("BOTTOMRIGHT")
end

local function OnDragHandleMouseUp(this, button)
	this.windowframe:StopMovingOrSizing()
	this.windowframe.window:OnSized()
	this.windowframe.isResizing = nil
end

local function OnDragHandleEnter(this)
	local hideEvent = this.hideEvent
	this.hideEvent = nil
	if this.windowframe then
		hideEvent = this.windowframe.hideEvent
		this.windowframe.hideEvent = nil
	end
	Violation:CancelScheduledEvent(hideEvent)
end

----------------------------
-- Bar Tabler helpers
----------------------------

local VioWindow -- early declaration

local currentTabletBar

local function UpdateBarTooltip()
	GameTooltip:SetOwner(currentTabletBar)

	local char = currentTabletBar.vioName
	GameTooltip:ClearLines()
	GameTooltip:AddLine(char)
	GameTooltip:AddLine(" ")
	GameTooltip:AddDoubleLine("Module", "Value", 1, 1, 1, 1, 1, 1)
	for index, modName in Violation:IterateRunningModules() do
		local displayName = runningModules[modName]:GetDisplayName()
		local value = Violation:GetModuleData(modName, char)
		if value and tonumber(value) > 0 then
			local formatter = Violation.dataFormatter[modName] or tostring
			local color = db.colors[modName]
			GameTooltip:AddDoubleLine(displayName, formatter(value), color.r, color.g, color.b, 1, 1, 1)
		end
	end
	GameTooltip:AddLine(" ")
	local source = select(2, Violation:GetUserDataSourceDesc(char))
	GameTooltip:AddDoubleLine(L["Data source"], source, 1, 1, 1, 0.2, 1, 0.2)
	GameTooltip:Show()
end

function OnBarEnter(this)
	if db.windowTooltips and not UnitAffectingCombat("player") or IsControlKeyDown() then
		currentTabletBar = this
		UpdateBarTooltip()
		Violation:ScheduleRepeatingEvent('Violation_UpdateBarTooltip', UpdateBarTooltip, 1)
	end
end

function OnBarLeave(this)
	if currentTabletBar == this then
		currentTabletBar = nil
		Violation:CancelScheduledEvent('Violation_UpdateBarTooltip')
		GameTooltip:ClearLines()
		GameTooltip:Hide()
	end
end

----------------------------
-- Violation Window Class
----------------------------

VioWindow = AceOO.Class()

Violation.VioWindow = VioWindow

function VioWindow.prototype:init()
	VioWindow.super.prototype.init(self)

	local f = CreateFrame("Frame", nil, UIParent)
	f:Hide()
	f.window = self
	self.frame = f
end

local windowBackdrop = {
	bgFile = "Interface\\Tooltips\\UI-Tooltip-Background", tile = true, tileSize = 16,
	edgeFile = "Interface\\AddOns\\Violation\\textures\\otravi-semi-full-border", edgeSize = 32,
	insets = {left = 0, right = 0, top = 20, bottom = 0},
}
function VioWindow.prototype:Setup()

	self.deleted = false

	local f = self.frame
	local o = self.options
	local font = GameFontNormal:GetFont()

	self.barOffset = 0
	if not self.currentBars then
		self.currentBars = {}
	end

	self:SetupMenuOptions()

	f:SetScale(o.scale or 1)
	f:SetClampedToScreen(true)

	f:SetWidth(o.width)
	f:SetHeight(o.height)
	f:SetPoint("CENTER", UIParent, "CENTER")
	f:SetBackdrop(windowBackdrop)
	f:SetBackdropColor(24/255, 24/255, 24/255)
	f:EnableMouse(true)
	f:EnableMouseWheel(true)

	f:SetScript("OnMouseUp", OnWindowMouseUp)
	f:SetScript("OnMouseWheel", OnWindowMouseWheel)

	-- header
	if not f.header then
		f.header = f:CreateFontString(nil,"OVERLAY")
	end
	f.header:SetFont(font, 12)
	f.header:SetText("")

	local module = Violation:HasModule(options.module) and Violation:GetModule(options.module) or nil

	if module then f.header:SetText(module:GetDisplayName()) end
	f.header:SetJustifyH("LEFT")
	f.header:SetShadowOffset(.8, -.8)
	f.header:SetShadowColor(0, 0, 0, 1)
	f.header:ClearAllPoints()
	f.header:SetPoint("TOPLEFT", f, "TOPLEFT", 24, -14)
	f.header:SetPoint("TOPRIGHT", f, "TOPRIGHT", -42, -14)

	-- close button
	if not f.closebutton then
		f.closebutton = CreateFrame("Button", nil, f)
	end
	f.closebutton:SetWidth(16)
	f.closebutton:SetHeight(16)
	f.closebutton:SetPoint("TOPRIGHT", f, "TOPRIGHT", -7, -13)
	f.closebutton:SetScript("OnClick", OnWindowClose)
	if not f.close then
		f.close = f:CreateTexture(nil, "ARTWORK")
	end
	f.close:SetPoint("CENTER", f.closebutton, "CENTER")
	f.close:SetTexture("Interface\\AddOns\\Violation\\textures\\otravi-close")
	f.close:SetWidth(16)
	f.close:SetHeight(16)
	f.close:SetBlendMode("ADD")

	-- prev button
	if not f.prevbutton then
		f.prevbutton = CreateFrame("Button", nil, f)
	end
	f.prevbutton:SetWidth(16)
	f.prevbutton:SetHeight(16)
	f.prevbutton:SetPoint("TOPLEFT", f, "TOPLEFT", 7, -13)
	f.prevbutton:SetScript("OnClick", OnWindowPrevModule)
	if not f.prev then
		f.prev = f:CreateTexture(nil, "ARTWORK")
	end
	f.prev:SetPoint("CENTER", f.prevbutton, "CENTER")
	f.prev:SetTexture("Interface\\AddOns\\Violation\\textures\\prev")
	f.prev:SetWidth(16)
	f.prev:SetHeight(16)
	f.prev:SetBlendMode("ADD")

	-- next button
	if not f.nextbutton then
		f.nextbutton = CreateFrame("Button", nil, f)
	end
	f.nextbutton:SetWidth(16)
	f.nextbutton:SetHeight(16)
	f.nextbutton:SetPoint("TOPRIGHT", f, "TOPRIGHT", -28, -13)
	f.nextbutton:SetScript("OnClick", OnWindowNextModule)
	if not f.next then
		f.next = f:CreateTexture(nil, "ARTWORK")
	end
	f.next:SetPoint("CENTER", f.nextbutton, "CENTER")
	f.next:SetTexture("Interface\\AddOns\\Violation\\textures\\next")
	f.next:SetWidth(16)
	f.next:SetHeight(16)
	f.next:SetBlendMode("ADD")

	-- drag handle
	if not f.draghandle then
		f.draghandle = CreateFrame("Frame", nil, f)
	end
	f.draghandle.windowframe = f
	f.draghandle:Hide()
	f.draghandle:SetFrameLevel(f:GetFrameLevel() + 10) -- place this above everything
	f.draghandle:SetWidth(16)
	f.draghandle:SetHeight(16)
	f.draghandle:SetPoint("BOTTOMRIGHT", f, "BOTTOMRIGHT", -1, 1)
	f.draghandle:EnableMouse(true)

	if not f.draghandle.texture then
		f.draghandle.texture = f.draghandle:CreateTexture(nil,"BACKGROUND")
	end
	f.draghandle.texture:SetTexture("Interface\\AddOns\\Violation\\textures\\draghandle")
	f.draghandle.texture:SetWidth(16)
	f.draghandle.texture:SetHeight(16)
	f.draghandle.texture:SetBlendMode("ADD")
	f.draghandle.texture:SetPoint("CENTER", f.draghandle, "CENTER", 0, 0)

	self:ToggleLock(o.locked)
	self:RestorePosition()

	self:Update()

	if o.show and db.showWindows then
		self:Open()
		Violation:ToggleWindows(true)
	else
		self:Close()
	end
end

function VioWindow.prototype:GetID()
	for k, w in ipairs(Violation.windows) do
		if w == self then
			return k
		end
	end
end

function VioWindow.prototype:DeleteWindowOptions()
	local id = self:GetID()
	if self.menuoptions then
		self.menuoptions.args = nil	-- this one is shared, don't nuke
		self.menuoptions.handler = nil	-- this one points to self == infinite recursion
		self.menuoptions = del(self.menuoptions)
	end
	options.args.windows.args[id] = nil
end

function VioWindow.prototype:UpdateOptionsName()
	local id = self:GetID()
	local mod = Violation:HasModule(self.options.module) and Violation:GetModule(self.options.module) or nil
	local modName = "?"
	if mod then
		modName = mod:GetDisplayName()
	end
	self.menuoptions.name = ("%d. %s"):format(id, modName)
	self.menuoptions.desc = L["Options for Window %d. %s"]:format(id, modName)
end

function VioWindow.prototype:SetupMenuOptions()
	local id = self:GetID()
	self:DeleteWindowOptions()
	self.menuoptions = newkeyed("type", "group", "handler", self, "args", windowOptionsTemplate,
		"isChecked", "windowIsShown", "onClick", "windowToggleShown")
	self:UpdateOptionsName()
	options.args.windows.args[id] = self.menuoptions
end

function VioWindow.prototype:OpenMenu()
	dew:Open(self.frame, "children", function() dew:FeedAceOptionsTable(self.menuoptions) end, "cursorX", true, "cursorY", true)
end

function VioWindow.prototype:ToggleLock(locked)
	self.options.locked = locked
	local f = self.frame
	if locked then
		f:SetMovable(nil)
		f:SetResizable(nil)
		f:RegisterForDrag()
		f.closebutton:Hide()
		f.close:Hide()

		f:SetScript("OnEnter", nil)
		f:SetScript("OnLeave", nil)
		f:SetScript("OnDragStart", nil)
		f:SetScript("OnDragStop", nil)
		f:SetScript("OnSizeChanged", nil)
		f.draghandle:SetScript("OnMouseDown", nil)
		f.draghandle:SetScript("OnMouseUp", nil)
		f.draghandle:SetScript("OnEnter", nil)
		f.draghandle:SetScript("OnLeave", nil)
	else
		f:SetMovable(true)
		f:SetResizable(true)
		f:SetMinResize(50,50)
		f:RegisterForDrag("LeftButton")
		f.closebutton:Show()
		f.close:Show()

		f:SetScript("OnDragStart", f.StartMoving)
		f:SetScript("OnDragStop", OnWindowDragStop)
		f:SetScript("OnSizeChanged", OnWindowSizeChanged)
		f:SetScript("OnEnter", OnDragHandleShow)
		f:SetScript("OnLeave", ScheduleDragHandleHide)
		f.draghandle:SetScript("OnMouseDown", OnDragHandleMouseDown)
		f.draghandle:SetScript("OnMouseUp", OnDragHandleMouseUp)
		f.draghandle:SetScript("OnEnter", OnDragHandleEnter)
		f.draghandle:SetScript("OnLeave", ScheduleDragHandleHide)
	end

end

function VioWindow.prototype:RestorePosition()
	local f = self.frame
	local x = self.options.posX
	local y = self.options.posY
	if x and y then
		local s = f:GetEffectiveScale()
		f:ClearAllPoints()
		f:SetPoint("TOPLEFT", UIParent, "BOTTOMLEFT", x / s, y / s)
	end
end

function VioWindow.prototype:OnOffset(up)
	-- we do not check if ouyr barOffset is still valid here. Update will do that for us
	if up then
		self.barOffset = self.barOffset - 1
	else
		self.barOffset = self.barOffset + 1
	end
	self:Update()
end

function VioWindow.prototype:OnSized()
	local f = self.frame
	local o = self.options

	if not f.isResizing then return end -- don't do anything on frame creation

	local y = f:GetHeight()
	local x = f:GetWidth()

	o.height = y
	o.width = x

	y = math.ceil(y)
	y = y - 34 -- deduct the header size

	local nrbars = math.floor(y / 16)
	if nrbars < 1 then nrbars = 1 end

	if nrbars ~= o.nrBars then
		if nrbars < o.nrBars then
			for i = nrbars+1, o.nrBars do
				self:ClearBar(i)
			end
		end
		o.nrBars = nrbars
		self:Update()
	else
		self:Update(true)
	end
end

function VioWindow.prototype:SwitchView(forward)
	local o = self.options
	local mod = o.module
	local dirtpile = new()
	local current
	for index, modName in Violation:IterateRunningModules() do
		table_insert(dirtpile, modName)
		if modName == mod then
			current = index
		end
	end
	if not current or (forward and current == #dirtpile) then
		o.module = dirtpile[1]
	elseif forward then
		o.module = dirtpile[current + 1]
	elseif current == 1 then
		o.module = dirtpile[#dirtpile]
	else
		o.module = dirtpile[current -1]
	end
	dirtpile = del(dirtpile)
	if mod ~= o.module then
		self:UpdateOptionsName()
		self:Update()
	end
end

function VioWindow.prototype:Delete()
	if dew:IsOpen(self.frame) then
		dew:Close()
	end
	self:Close()
	self:DeleteWindowOptions()
end

function VioWindow.prototype:Close()
	self.frame:Hide()
	for k, v in ipairs(self.currentBars) do
		self:ClearBar(k)
	end
end

function VioWindow.prototype:Open()
	self.frame:Show()
	self:Update()
end

function VioWindow.prototype:SetBar(count, rank, name, value, percent)
	local texture = self.options.useDefaultTexture and db.defaultBarTexture or self.options.barTexture

	if not self.currentBars[count] then
		self.currentBars[count] = NewBar(texture)
	end
	local f = self.currentBars[count]

	f:SetParent(self.frame)
	f.vioWindow = self
	f.vioName = name

	f.lefttext:SetText(self.options.useSimpleBars and name or format("%s %s", rank, name))

	local textValue = dataInfo.formatter(value)

	if percent and not self.options.useSimpleBars then
		f.righttext:SetText(format("%s %s", textValue, percentString:format(percent)))
	else
		f.righttext:SetText(textValue)
	end

	local vmax = dataInfo.maxValue

	if vmax and vmax ~= 0 then
		f.statusbar:SetValue((value/vmax)*100)
	else
		f.statusbar:SetValue(100)
	end
	local color = RGB[name]
	if not color then
		f.statusbar:SetStatusBarColor(.5, .5, .5, 1)
	else
		f.statusbar:SetStatusBarColor(color[1], color[2], color[3], 1)
	end
	if count == 1 then
		f:SetPoint("TOP", self.frame, "TOP", 0, -32)
	else
		f:SetPoint("TOP", self.currentBars[count-1], "BOTTOM", 0, -1)
	end
	f.statusbar:SetStatusBarTexture(SM:Fetch("statusbar", texture))
	f:SetPoint("LEFT", self.frame, "LEFT", 0, 0)
	f:SetPoint("RIGHT", self.frame, "RIGHT", 0, 0)
	f:Show()
end

function VioWindow.prototype:ClearBar(count)
	if not self.currentBars[count] then return end
	self.currentBars[count] = FreeBar(self.currentBars[count])
end

function VioWindow.prototype:UpdateIfDirty()
	local mod = Violation:HasModule(self.options.module) and Violation:GetModule(self.options.module) or nil
	if mod and mod:IsDirty() then
		self:Update()
		return self.options.module
	end
end

local disabledWindowColor = {r=0.5, g=0.5, b=0.5, a=1}

function VioWindow.prototype:Update(sameamount)
	if not self.options.show then return end

	local o = self.options

	local mod = o.module
	local module = Violation:HasModule(mod) and Violation:GetModule(mod) or nil
	if module then
		o.cachedShortDisplayName = module:GetShortDisplayName()
		o.cachedDisplayName = module:GetDisplayName()
	end

	local f = self.frame

	local txt
	if (f.header:GetRight() or 0) - (f.header:GetLeft() or 0) < 80 then
		txt = o.cachedShortDisplayName or "?"
	else
		txt = o.cachedDisplayName or "?"
	end

	if not db.enabledModules[mod] then
		module = nil
		txt = format("%s\n\n%s\nnot active",txt,txt)	-- uuuugly
	end

	f.header:SetText(txt)

	local color
	if not module then
		color = disabledWindowColor
	else
	 	color = o.useModuleColor and db.colors[mod] or o.color
	end
	f:SetBackdropBorderColor(color.r, color.g, color.b)

	if f.barScale ~= (o.scale or 1) then
		local x = f:GetLeft()
		local y = f:GetTop()
		local scale = f:GetScale()
		x = x * scale
		y = y * scale
		scale = o.scale or 1
		f:SetScale(scale)
		x = x / scale
		y = y / scale
		f:ClearAllPoints()
		f:SetPoint("TOPLEFT", UIParent, "BOTTOMLEFT", x,y)
		f.barScale = o.scale
	end

	if module then
		if not sameamount then
			local data = Violation:GetModuleData(module.name)
			if type(data) == "table" then
				self:UpdateBars(module.name)
			end
		end
	else
		for k, v in ipairs(self.currentBars) do
			self:ClearBar(k)
		end
	end
end

function VioWindow.prototype:UpdateBars(modName)
	local o = self.options
	local barCount, barOff = Violation:PublishModuleData(modName, self.BarAddCallBack, o.nrBars, self.barOffset, self)
	self.barOffset = barOff -- set our barOffset to the correct new value.
	if barCount < o.nrBars then
		for i = barCount+1, o.nrBars do
			self:ClearBar(i)
		end
	end
end

function VioWindow.prototype:BarAddCallBack(i, rank, name, value, percentage)
	return self:SetBar(i, rank..".", name, value, percentage)
end

-- aceoptions handler methods:
function VioWindow.prototype:windowGet(passValue)
	return self.options[passValue]
end

function VioWindow.prototype:windowSet(passValue, value)
	self.options[passValue] = value
	self:Setup()
end

function VioWindow.prototype:windowSetSoft(passValue, value)
	self.options[passValue] = value
	self:Update(true)
end

function VioWindow.prototype:windowIsShown()
	return self.options["show"]
end

function VioWindow.prototype:windowToggleShown()
	self.options["show"] = not self.options["show"]
	self:Setup()
end

function VioWindow.prototype:windowDelete(passValue)
	Violation:DeleteWindow(self)
end

function VioWindow.prototype:windowReset(passValue)
	local mod = Violation:HasModule(self.options.module) and Violation:GetModule(self.options.module)
	if mod then
		mod:Reset()
		self:Update()
	end
end

function VioWindow.prototype:windowSetColor(passValue, r, g, b, a)
	local tuple = self.options[passValue]
	self.r = r
	self.g = g
	self.b = b
	self.a = a
	self:Update(true)
end

function VioWindow.prototype:windowGetColor(passValue)
	local tuple = self.options[passValue]
	return tuple.r, tuple.g, tuple.b, tuple.a
end

function VioWindow.prototype:windowUsesModuleColor()
	return self.options.useModuleColor
end

function VioWindow.prototype:windowUsesDefaultTexture()
	return self.options.useDefaultTexture
end

function VioWindow.prototype:onClickReporting(...)
	local mod = Violation:HasModule(self.options.module) and Violation:GetModule(self.options.module)
	return mod and mod:onClickReporting(...)
end

function VioWindow.prototype:getReporting(...)
	local mod = Violation:HasModule(self.options.module) and Violation:GetModule(self.options.module)
	return mod and mod:getReporting(...)
end

function VioWindow.prototype:setReporting(...)
	local mod = Violation:HasModule(self.options.module) and Violation:GetModule(self.options.module)
	return mod and mod:setReporting(...)
end

